/***************************************************************************
 *   Copyright (C) 2005 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef ANALYZER_H
#define ANALYZER_H

#include "ScaleDeltaFilter.h"
#include "DecompositionLevel.h"

/// Final structure is a vector of data + its decomposition
/// Name based on Dynamic Multi-Fractal Analysis with Discrete Wavelet Transform
/// Suggested names like DMFADWT, diwtradymfrana... Too complex
/// Let's name it something simple
class Analyzer {
protected:

    /// Share common computations with other analyzers
    ScaleDeltaFilterArray* scaleDeltaFilterArray;

    /// Cache to the real filter array
    ScaleDeltaFilter* scaleDeltaFilter;

    /// This will be set to 2^L. This is also the number fo filters
    int nframes;
    double nframesInv;
    /// binary mask for the filters. Well be set to 2^L - 1
    int filter_index_mask;

    // pointer to member type
    typedef DecompositionLevel* (Analyzer::* NewDataDWTFunctionPtr)(double);
    NewDataDWTFunctionPtr newDataDWTPtr;

    /// Updates the DWT decomposition, takes care of non-full buffer effects
    /// and automatically switched to the full buffer version when this happens
    DecompositionLevel* newDataDWTInitPhase(double newData);

    /// Faster version of the DWT update, assuming the buffer is full
    DecompositionLevel* newDataDWTFullRegime(double newData);
    int newDataDWTIdx;

    int newMFAFrameFirstIdx;

    bool discardOlderData;
    bool averageAllFrames;
    bool monitorError;

    // theLogs is unused when not monitoring Errors
    // tmp variable otherwise
    double** theLogs;
    double sumOfWeightsInv;
    double* weights;
    double* invq;
    double* invs;

    /// Protected helper called by the constructors
    void constructor(int nlevel, int ncoefmin, Wavelet* w, double* weights, bool discardOlderData, bool averageAllFrames, bool monitorError);

    

public:
    /// These arrays are used only when averageAllFrames is true (the default)
    CyclicArrayTemplate<double*> hqHistory;

    /// These arrays are used only when both averageAllFrames & monitorError are true (not the default)
    CyclicArrayTemplate<double*> errorHistory;

    /// Scale dependent constant for computing H(q) from q-power sums
    /// array with q & scale as index. scale 0 (data) is not used
    /// See article about incremental MFA for doc.
    double **powerSumToHqConstant;

    /// Scale dependent constant for computing the b in y = b.s^(ax) weighted
    /// exponential fit from the a values (actually the log2 value of it)
    double **powerSumToExpFitResidualConstant;

    /// the h(q) values. The implementation in Analyzer.cpp uses a hardcoded convention.
    /// hq is an array [0..19] of 20 double. q=1..10 is stored first in hq[0..9],
    /// q=-1..-10 is stored in hq[10..19].
    /// See AnalyzerReducedNq.cpp for a convention for q=-5..5. Only Analyzer.cpp needs
    /// to be changed when the hq convention is changed, the other files are independent.
    double *hq;

    /// mean square error for the h(q) weighted least square estimate, averaged on all levels.
    /// Available only when the 'monitorError' flag was used.
    /// See hq for the array indices signification.
    double *meanSqError;

    int nlevel;

    /// In any method using a spectrum there is a delay to get low-frequency components.
    /// This is completely independent from this algorithm, but a general fact.
    /// In our case, the delay corresponds to the difference between the newly arrived point
    /// and the last point that can be reconstructed using the DWT.
    /// In practice, it is (2^L - 1)(wavelet->support - 2)
    int delay;

    DecompositionLevel *nextLevel;
    DecompositionLevel *nextLevelWaiting;

    /// Original data: the low pass for next level
    CyclicArray data;
    /// index of the last reconstructible point
    /// delay due to low pass frequency acquisition
    ///int delay;

    /// Analyzing wavelet
    Wavelet* wavelet;
    
    /** Public constructor takes these arguments:
        @param nlevel    Number of levels of decomposition
        @param ncoefmin  Size of the highest level of decomposition. For correct operation, this should be
                         at least wavelet->support - 1, that is, the maximum Scale Delta Filter size.
                         When keeping old data, this parameter has no effect on the result (hence the minimum is fine)
                         When discarding old date (non stationary series), set this parameter to some significant value
                         to make statistics on (10 may not be enough for example).
        @param w         The wavelet
        @param discardOlderData   Whether older data should be discarded or not. Default is true
        @param averageAllFrames   Whether to enable frame averaging or not (recommended, see article). Default is true

        @note  The data size is 2^nlevel * ncoefmin + (w->support - 2) * (2^nlevel - 1), with ^ = exponentiation
    */
    Analyzer(int nlevel, int ncoefmin, Wavelet* w, bool discardOlderData = true, bool averageAllFrames = true, bool monitorError = false);

    Analyzer(int nlevel, int ncoefmin, Wavelet* w, double* weights, bool discardOlderData = true, bool averageAllFrames = true, bool monitorError = false);

    ~Analyzer();

    /// Discrete Wavelet Transform the data into the next level
    /// low and high band coefficients
    void transform(CyclicArray& data, DecompositionLevel* level);

    /// restore data from next level by inverse DWT
    void restore(DecompositionLevel* level, CyclicArray& data);

    /// Algo from fast MFA-update paper. See this ref.
    double computeScaleDelta(int dataIndex, int scale, CyclicArray& hp);

    /// Update DWT for this new data.
    inline DecompositionLevel* newDataDWT(double newData) {
        return (this->*newDataDWTPtr)(newData);
    }

    /// Update MFA for this new data. Calls update DWT internally
    /// Returns 2 so long as there is not enough data (initialization
    /// phase). Then returns 1 so long as there are non-incremental version
    /// called to end the initialization phase (2^L such calls). Then
    /// returns 0. When it returns 0 or 1, the h(q) values have been updated.
    int newDataMFA(double data);

    /// Non-incremental O(N) version of the spectrum computations.
    /// This is used during the initialization phase, but can also be used
    /// for consistency checks.
    /// Note1: Incremental updates proceed by adding and substracting terms
    ///        in power sums. After a lot of updates, this may possibly lead
    ///        to slight numerical roundup errors. In that case, computing
    ///        directly the full spectrum resets the sums to their exact
    ///        values.
    /// Note2: This function is called as soon as possible by the initialization
    ///        process, there is no point in calling it before (and it would lead
    ///        to incorrect computations anyway). Partial results could however
    ///        be computed with care using the incremental code. This is not
    ///        handled: waiting for the full spectrum is preferred over taking
    ///        care of all the computations for partial results. See the code
    ///        comments in the incremental version if you want to implement that.
    void nonIncrementalMFA();
};

#endif
