/***************************************************************************
 *   Copyright (C) 2005 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef DECOMPOSITION_LEVEL_H
#define DECOMPOSITION_LEVEL_H

#include "CyclicArray.h"
#include "Wavelet.h"

/** Fast DWT update algo is based on recursive half-window overlapping levels
    One of these is the active one, corresponding to the even data frames
    The other one has the last data point waiting around. When a new point arrives,
    it completes this odd frame into an even one, and the buffers are swapped.
    The new point was also added to the previously active buffer, which becomes odd-framed
    with the last point in the waiting state.

    Note this doesn't introduce an additional delay in the DWT coeficients: the last
    point is always taken in account in the new "active" buffer. Of course, the normal delay
    applies for getting low-frequency components out of an instantaneous measure, but
    this is completely unrelated to this algorithm. Recursively, a whole new DWT tree of active
    buffers is updated in O(L) constant time (it does not depend on the data buffer size).
    Another way to look at this algorithm is to consider there are 2^L windows maintained,
    in parallel, ending at the last 2^L consecutive indices. However, most of these frames
    share common points: for example two consecutive even frames of the first decomposition
    of the data have N/2 points each, among which N/2 - 1 are common points.

    The buffer pointers are kept around for easy pointer swapping on data update. Therefore,
    nextLevel and nextLevelHP always point to the current, valid, decomposition.
*/
class DecompositionLevel {
protected:

    // See class description
    DecompositionLevel* nextLevelWaiting;

    int newDataUpdateIdx;

    // pointer to member type
    typedef DecompositionLevel* (DecompositionLevel::* NewDataFunctionPtr)(double*); 
    NewDataFunctionPtr newDataUpdatePtr;

    /** Updates this buffer with new data. As this is a cyclic buffer, the oldest one is discarded.
        The waiting and active next level buffers are updated & swapped as necessary.
        This operation is constant-time O(L) with L number of decomposition levels.
        Warning: this version is designed for better performance and assumes the array is
        full (use the other method for the initialization phase).
     */
    DecompositionLevel* newDataUpdateFullRegime(double* lphp);

    /** Updates this buffer with new data. As this is a cyclic buffer, the oldest one is discarded.
        The waiting and active next level buffers are updated & swapped as necessary.
        This operation is constant-time O(L) with L number of decomposition levels
        This method could be used in both Initialization phase and for the full regime,
        but the other method will be more efficient for the full regime. Hence
        this function automatically switches to the other one when ready
     */
    DecompositionLevel* newDataUpdateInitPhase(double* lphp);

    // common wavelet instance shared amongst all decomposition levels
    Wavelet* wavelet;

public:

    /** Create a tree of depth nlevel, with ncoefmin coefficients in
        the highest level low-pass filer (that is, the level with
        larger scale and thus lesser number of coefficients)
        Each level uses up ws extra coefficients where ws is the wavelet
        support size: ws=4 for Daubechies W4 for example.

        This is due to the fact we don't wan't to make the data cyclic
        or extend the last values. This would create harmful artifacts, and
        complicate things in our case of cyclic buffers.

        However, for our usage, the data is actually a history array, and we
        don't care if the oldest values are not reconstructed. On the other
        hand, that half the discarded values are new ones may be problemetic
        for the delay introduced between the arrival of a new data and it's full
        inclusion in the spectrum computations.

        Computes capacity before recursion (alternative: use a class factory)
    */
    DecompositionLevel(int nlevel, int ncoefmin, Wavelet* w);
    ~DecompositionLevel();

    /// next level data.
    DecompositionLevel* nextLevel;

    /// The coefficients
    CyclicArray lowPass;
    CyclicArray highPass;

    /** Updates this buffer with new data. As this is a cyclic buffer, the oldest one is discarded.
        The waiting and active next level buffers are updated & swapped as necessary.
        This operation is constant-time O(L) with L number of decomposition levels.
     */
    inline DecompositionLevel* newDataUpdate(double* lphp) {
        return (this->*newDataUpdatePtr)(lphp);
    }

    /// Power sum of the delta at this scale for different q (double array[n_q])
    double *powerSum;

/*
    /// update the delta values for this scale and power series, then pass array up to next level.
    /// According to user choice the first (2^L - 2^s) values can be recomputed (uses CPU,
    /// no mem) or re-used from last update (uses less CPU, but this quickly uses a lot of
    /// memory since there are 2^s decomposition levels at scale s!)
    void updateScaleDelta(int dataidx, int scale, double* delta);
*/
};

/* TODO
/// Add a memory cache to speed up the delta computations, at the expense
/// of higher memory consumption
class DeltaCachingDecompositionLevel : public DecompositionLevel {
public:

    DeltaCachingDecompositionLevel(int nlevel, int ncoefmin, Wavelet* w, int maxq, int scale) :
    DecompositionLevel(nlevel, ncoefmin, w, maxq, scale) {}

    /// Values that can be re-used instead of recomputed
    /// (2^L - 2^s) such values are present.
    /// Warning: this quickly uses a lot of memory since there are 2^s decomposition
    /// levels at scale s!
    double *previousDeltaValues;
    /// Number of such values : 2^L - 2^s, computed in constructor
    int numSharedScaleDeltas;

    /// Variant of parent function using more memory and less CPU
    virtual void updateScaleDelta(int dataidx, int scale, double* delta);
};
*/

#endif
