/***************************************************************************
 *   Copyright (C) 2005 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "Utility.h"
#include <iostream>
using namespace std;
#include <time.h>

#include <math.h>

/*

==============================================================================

The following class definition is taken from "randomc.h", a C++ random number
generator library released uner the GNU GPL by Agner Fog.

See http://www.agner.org/random/randomc.htm for more information and the
original files.

This derived work contains only the definition of the Mersenne twister
random number generator, and was adapted to:
- Remove unused constants
- TODO: Made architecture detection from configure script
*/

// Define 32 bit signed and unsigned integers.
// Change these definitions, if necessary, on 64 bit computers
typedef   signed long int32;     
typedef unsigned long uint32;     

class TRandomMersenne {                // encapsulate random number generator
    // constants for MT19937:
    #define MERS_N   624
    #define MERS_M   397
    #define MERS_R   31
    #define MERS_U   11
    #define MERS_S   7
    #define MERS_T   15
    #define MERS_L   18
    #define MERS_A   0x9908B0DF
    #define MERS_B   0x9D2C5680
    #define MERS_C   0xEFC60000
public:
    TRandomMersenne(uint32 seed) {       // constructor
        RandomInit(seed);}

    void RandomInit(uint32 seed);        // re-seed
    void RandomInitByArray(uint32 seeds[], int length); // seed by more than 32 bits
    int IRandom(int min, int max);       // output random integer
    double Random();                     // output random float
    uint32 BRandom();                    // output random bits

private:
    uint32 mt[MERS_N];                   // state vector
    int mti;                             // index into mt

// win32 problem: defines in winsock2.h prevent correct enum
// unconditional undef for this cpp file as we don't use these constants here anyway
#undef LITTLEENDIAN
#undef BIGENDIAN
#undef NONIEEE
    
    enum TArch {LITTLEENDIAN, BIGENDIAN, NONIEEE};
    TArch Architecture;                  // conversion to float depends on computer architecture
};    

#include "mersenne.cpp"

/* 
End of randomc library re-use
==============================================================================
*/




namespace Utility {

// start without a generator and allocate it on demand
// => avoid creating a RNG with time seed and re-seeding
static TRandomMersenne* rng = 0;

static long randomSeed = 42;

static inline TRandomMersenne* getRNG() {
    if (rng!=0) return rng;
    RandomInit();
    return rng;
}

}

/*
    patest_pink.c

    Generate Pink Noise using Gardner method.
    Optimization suggested by James McCartney uses a tree
    to select which random value to replace.

    x x x x x x x x x x x x x x x x 
     x   x   x   x   x   x   x   x   
       x       x       x       x       
           x               x               
                   x                               

    Tree is generated by counting trailing zeros in an increasing index.
    When the index is zero, no random number is selected.

    Author: Phil Burk, http://www.softsynth.com
    
    Revision History:

    Copyleft 1999 Phil Burk - No rights reserved.
*/


#define PINK_MAX_RANDOM_ROWS   (30)
#define PINK_RANDOM_BITS       (24)
#define PINK_RANDOM_SHIFT      ((sizeof(long)*8)-PINK_RANDOM_BITS)

typedef struct
{
    long      pink_Rows[PINK_MAX_RANDOM_ROWS];
    long      pink_RunningSum;   /* Used to optimize summing of generators. */
    int       pink_Index;        /* Incremented each sample. */
    int       pink_IndexMask;    /* Index wrapped by ANDing with this mask. */
    double    pink_Scalar;       /* Used to scale within range of -1.0 to +1.0 */
} PinkNoise;

/* Setup PinkNoise structure for N rows of generators. */
void InitializePinkNoise( PinkNoise *pink, int numRows )
{
    int i;
    long pmax;
    pink->pink_Index = 0;
    pink->pink_IndexMask = (1<<numRows) - 1;
/* Calculate maximum possible signed random value. Extra 1 for white noise always added. */
    pmax = (numRows + 1) * (1<<(PINK_RANDOM_BITS-1));
    pink->pink_Scalar = 1.0f / pmax;
/* Initialize rows. */
    for( i=0; i<numRows; i++ ) pink->pink_Rows[i] = 0;
    pink->pink_RunningSum = 0;
}

/* Generate Pink noise values between -1.0 and +1.0 */
double GeneratePinkNoise( PinkNoise *pink )
{
    long newRandom;
    long sum;
    double output;

/* Increment and mask index. */
    pink->pink_Index = (pink->pink_Index + 1) & pink->pink_IndexMask;

/* If index is zero, don't update any random values. */
    if( pink->pink_Index != 0 )
    {
    /* Determine how many trailing zeros in PinkIndex. */
    /* This algorithm will hang if n==0 so test first. */
        int numZeros = 0;
        int n = pink->pink_Index;
        while( (n & 1) == 0 )
        {
            n = n >> 1;
            numZeros++;
        }

    /* Replace the indexed ROWS random value.
     * Subtract and add back to RunningSum instead of adding all the random
     * values together. Only one changes each time.
     */
        pink->pink_RunningSum -= pink->pink_Rows[numZeros];
        newRandom = ((long)Utility::getRNG()->BRandom()) >> PINK_RANDOM_SHIFT;
        pink->pink_RunningSum += newRandom;
        pink->pink_Rows[numZeros] = newRandom;
    }
    
/* Add extra white noise value. */
    newRandom = ((long)Utility::getRNG()->BRandom()) >> PINK_RANDOM_SHIFT;
    sum = pink->pink_RunningSum + newRandom;

/* Scale to range of -1.0 to 0.9999. */
    output = pink->pink_Scalar * sum;

    return output;
}


namespace Utility {

// Uninitialized pink noise
static PinkNoise thePinkNoise;

long RandomInit() {
    randomSeed = time(0);
    if (rng==0) rng = new TRandomMersenne(randomSeed);
    else rng->RandomInit(randomSeed);
    PinkNoiseSetRows();
    cerr << "Random seed=" << randomSeed << endl;
    return randomSeed;
}

void RandomInit(long seed) {
    randomSeed = seed;
    if (rng==0) rng = new TRandomMersenne(randomSeed);
    else rng->RandomInit(randomSeed);
    PinkNoiseSetRows();
    cerr << "Random seed=" << randomSeed << endl;
}

void PinkNoiseSetRows(int nrows) {
    InitializePinkNoise(&thePinkNoise, nrows);
}

int IRandom(int min, int max) {
    return getRNG()->IRandom(min, max);
}

double Random() {
    return getRNG()->Random();
}

unsigned long BRandom() {
    return getRNG()->BRandom();
}

#define TWO_PI (2.0*3.1415926535897932384626433832795029)

double NRandom() {
    double x;
    do {
        x = Random();
    } while (x==0.0);
    x = sqrt(-2.0 * log(x));
    if (Random() > 0.5) return x * cos(TWO_PI * Random());
    return x * sin(TWO_PI * Random());
}

double PRandom() {
    return (GeneratePinkNoise(&thePinkNoise)+1.0)*0.5;
}

double b0=0.0, b1=0.0, b2=0.0, b3=0.0, b4=0.0, b5=0.0, b6=0.0;
double PRandom2() {
    double white = Random();
    b0 = 0.99886 * b0 + white * 0.0555179;
    b1 = 0.99332 * b1 + white * 0.0750759;
    b2 = 0.96900 * b2 + white * 0.1538520;
    b3 = 0.86650 * b3 + white * 0.3104856;
    b4 = 0.55000 * b4 + white * 0.5329522;
    b5 = -0.7616 * b5 - white * 0.0168980;
    double pink = b0 + b1 + b2 + b3 + b4 + b5 + b6 + white * 0.5362;
    b6 = white * 0.115926;
    return pink;
}

} // end namespace

