/***************************************************************************
 *   Copyright (C) 2005 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <iostream>
#include <fstream>
#include <sstream>
using namespace std;

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "Analyzer.h"
#include "Utility.h"

#define HAVE_LINUX_SYSCALL
#include "timeMeasurement.h"

int main(int argc, const char** argv) {

    Wavelet* wavelet = DW6::getInstance();
    int max_level = 10;
    int max_nmin = 50;
    int nmeasures = 3 * 3; // number of level loops * number of nmin loops
    const int Nruns = 30;

    double* avgNonIncr = new double[nmeasures];
    memset(avgNonIncr, 0, nmeasures*sizeof(double));
    double* devNonIncr = new double[nmeasures];
    memset(devNonIncr, 0, nmeasures*sizeof(double));
    double* avgIncr = new double[nmeasures];
    memset(avgIncr, 0, nmeasures*sizeof(double));
    double* devIncr = new double[nmeasures];
    memset(devIncr, 0, nmeasures*sizeof(double));
    
    
    // Use low and high bits for the seed
    for (int run =0; run < Nruns; ++run) {
        Utility::RandomInit(1L<<run);
        int measureIdx = 0;
        
        // number of levels of decomposition.
        for (int nlevel = 4; nlevel <= max_level; nlevel+=3) {
        
            // pre-allocate filter for all loops using this level
            ScaleDeltaFilterArray* filters = ScaleDeltaFilterArray::getInstance(wavelet, 5);
    
            // Number of data at the highest level
            for (int nmin = 10; nmin<=max_nmin; nmin += 20) {
    
                cout << "Processing nlevel=" << nlevel << ", nmin=" << nmin << endl;
                Analyzer analyzer(nlevel, nmin, wavelet);
                
                // Integrated white noise
                double iwn_value = 0.0;
            
                // Generate the series before the perf measure so it does not interfere
                // Use (1 << nlevel) non-incremental and 2*(1 << nlevel) incremental updates to average on
                int max_points = analyzer.data.capacity() + (1 << nlevel)*3;
            
                // Forward rectangular integration of white noise is Brownian motion
                double *iwn = new double[max_points];
                iwn[0] = 0.0;
                for (int i=1; i<max_points; ++i) {
                    iwn[i] = iwn[i-1] + Utility::NRandom();
                }
            
                // skip init phase
                int nextStage = analyzer.data.capacity()-1;
                int idx = 0;
                for (; idx<nextStage; ++idx) analyzer.newDataMFA(iwn[idx]);
            
                // There are exactly 2^L calls to the full MFA analysis. Use them for stats
                nextStage += 1 << nlevel;
                struct timespec ts;
                double timeSpent = -1.0;
                if (getCurrentTime(&ts)==0) timeSpent = ts.tv_sec + ts.tv_nsec * 1e-9;
                for (; idx<nextStage; ++idx) analyzer.newDataMFA(iwn[idx]);
                if (getCurrentTime(&ts)==0) timeSpent = (ts.tv_sec + ts.tv_nsec * 1e-9 - timeSpent) / (double)(1 << nlevel);
                cout << "Average non-incremental update time:" << (timeSpent*1e6) << " usec." << endl;
                avgNonIncr[measureIdx] += timeSpent;
                devNonIncr[measureIdx] += timeSpent*timeSpent;
            
                timeSpent = -1.0;
                if (getCurrentTime(&ts)==0) timeSpent = ts.tv_sec + ts.tv_nsec * 1e-9;
                // Now process timings experiment on incremental updates
                for (int i=idx; i<max_points; ++i) analyzer.newDataMFA(iwn[i]);
                // get time again
                if (getCurrentTime(&ts)==0) timeSpent = (ts.tv_sec + ts.tv_nsec * 1e-9 - timeSpent) / (double)(max_points - idx);
                cout << "Average incremental update time:" << (timeSpent*1e6) << " usec." << endl;
                avgIncr[measureIdx] += timeSpent;
                devIncr[measureIdx] += timeSpent*timeSpent;
            
                cout << endl;
                cout << "H(q) results (sanity check, should be monofractal H=0.5)." << endl;
                for (int i=0; i<5; ++i) cout << "h(" << (i+1) << ")=" << analyzer.hq[i] << " ";
                for (int i=5; i<10; ++i) cout << "h(" << (4-i) << ")=" << analyzer.hq[i] << " ";
                cout << endl;
                
                delete [] iwn;
                ++measureIdx;
            }
    
            filters->unref();
        }
    
    }

    //Final stats
    int measureIdx = 0;
    for (int nlevel = 4; nlevel <= max_level; nlevel+=3) {
        for (int nmin = 10; nmin<=max_nmin; nmin += 20) {
            cout << "Stats for nlevel=" << nlevel << ", nmin=" << nmin << endl;
            avgNonIncr[measureIdx] /= Nruns;
            devNonIncr[measureIdx] = sqrt(devNonIncr[measureIdx] / Nruns - avgNonIncr[measureIdx] * avgNonIncr[measureIdx]);
            avgIncr[measureIdx] /= Nruns;
            devIncr[measureIdx] = sqrt(devIncr[measureIdx] / Nruns - avgIncr[measureIdx] * avgIncr[measureIdx]);
            cout << "Overall non-incremental update time:" << avgNonIncr[measureIdx]*1e6 << " usec average, " << devNonIncr[measureIdx]*1e6 << " standard dev" << endl;
            cout << "Overall incremental update time:" << avgIncr[measureIdx]*1e6 << " usec average, " << devIncr[measureIdx]*1e6 << " standard dev" << endl;
            cout << "Ratio: " << (avgNonIncr[measureIdx] / avgIncr[measureIdx]) << endl;
            ++measureIdx;
        }
    }
    
}
