/***************************************************************************
 *   Copyright (C) 2005 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
using namespace std;

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "Analyzer.h"
#include "Utility.h"

int main(int argc, const char** argv) {
    Wavelet* wavelet = DW6::getInstance();
    
    // Use correct weights for exponential fitting of integrated pink noise
    // NOT using these weights would introduce a bias & loss of precision
    double weights[10];
    for (int s=1; s<=10; ++s) {
        weights[s-1] = (double)(1 << s);
    }

    // Also consider traditional unweighted least square
    double no_weights[10];
    for (int s=1; s<=10; ++s) no_weights[s-1] = 1.0;

    const int max_level = 10;
    const int max_nmin = 50;
    const int nmeasures = 3 * 3; // number of level loops * number of nmin loops

#define INITIALIZER \
{ \
{.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0}, \
{.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0}, \
{.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0}, \
{.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0}, \
{.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0}, \
{.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0}, \
{.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0}, \
{.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0}, \
{.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0,.0} \
}
    // Stats
    double avg_iwn_disc[nmeasures][20] = INITIALIZER;
    double dev_iwn_disc[nmeasures][20] = INITIALIZER;
    double avg_iwn_dev[nmeasures][20] = INITIALIZER;
    double dev_iwn_dev[nmeasures][20] = INITIALIZER;
    double avg_iwn_keep[nmeasures][20] = INITIALIZER;
    double dev_iwn_keep[nmeasures][20] = INITIALIZER;
    double avg_iwn_delta[nmeasures][20] = INITIALIZER;
    double dev_iwn_delta[nmeasures][20] = INITIALIZER;
    double avg_ipn_disc[nmeasures][20] = INITIALIZER;
    double dev_ipn_disc[nmeasures][20] = INITIALIZER;
    double avg_ipn_dev[nmeasures][20] = INITIALIZER;
    double dev_ipn_dev[nmeasures][20] = INITIALIZER;
    double avg_ipn_keep[nmeasures][20] = INITIALIZER;
    double dev_ipn_keep[nmeasures][20] = INITIALIZER;
    double avg_ipn_delta[nmeasures][20] = INITIALIZER;
    double dev_ipn_delta[nmeasures][20] = INITIALIZER;

    
    const int Nruns = 30;
    // Use low and high bits for the seed
    for (int run =0; run < Nruns; ++run) {
        Utility::RandomInit(1L<<run);
        int measureIdx = 0;
    
        // To be fair, process all configurations on the same series in each run.
    
        // Integrated white noise
        double iwn_value = 0.0;
        // Integrated pink noise
        double ipn_value = 0.0;
        
        // Generate the series
        // Ensure there is at least enough points at the max level
        int max_points = (1 << max_level) * max_nmin + ((1 << max_level) -1)* (wavelet->support - 2) + (1 << max_level)*3;
        
        // Forward rectangular integration of white noise is Brownian motion
        double *iwn = new double[max_points];
        double *ipn = new double[max_points];
        iwn[0] = 0.0;
        ipn[0] = 0.0;
        for (int i=1; i<max_points; ++i) {
            iwn[i] = iwn[i-1] + Utility::NRandom();
            ipn[i] = ipn[i-1] + Utility::PRandom();
        }
        
        // number of levels of decomposition.
        for (int nlevel = 4; nlevel <= max_level; nlevel+=3) {
        
            // pre-allocate filter for all loops using this level
            ScaleDeltaFilterArray* filters = ScaleDeltaFilterArray::getInstance(wavelet, 5);
    
            // Number of data at the highest level
            for (int nmin = 10; nmin<=max_nmin; nmin += 20) {
    
                // Create analyzers.
                // Process both "discard old points" and "keep old points" versions
                // For "discard old points", process both "average over 2^L frames"
                // and "do not average" so as to get standard deviation.
    
                // Analyzer for brownian motion (integrated white noise)
                // discard old points, no frame average
                Analyzer iwn_ana_disc_noavg(nlevel, nmin, wavelet, true, false);
                // Analyzer for integrated pink noise
                // discard old points, no frame average
                // Use correct weights
                Analyzer ipn_ana_disc_noavg(nlevel, nmin, wavelet, weights, true, false);
                // frame averaging versions
                Analyzer iwn_ana_disc_avg(nlevel, nmin, wavelet, true, true);
                Analyzer ipn_ana_disc_avg(nlevel, nmin, wavelet, weights, true, true);
    
                // Keep old points versions, frame average
                Analyzer iwn_ana_keep(nlevel, nmin, wavelet, false, true);
                Analyzer ipn_ana_keep(nlevel, nmin, wavelet, weights, false, true);
    
                // Keep old points versions, frame average, no weighting
                Analyzer iwn_ana_keep_no_weights(nlevel, nmin, wavelet, no_weights, false, true);
                Analyzer ipn_ana_keep_no_weights(nlevel, nmin, wavelet, no_weights, false, true);
    
                // Index of the first point that has a full history of frames for averaging
                int firstAvgFullHistory = iwn_ana_disc_avg.data.size() - 1 + (1<<nlevel);
                double iwn_frame_to_frame_dev[20];
                memset(iwn_frame_to_frame_dev, 0, 20*sizeof(double));
                double ipn_frame_to_frame_dev[20];
                memset(ipn_frame_to_frame_dev, 0, 20*sizeof(double));
    
                // Process the given series with all analyzers
                for (int i=0; i<max_points; ++i) {
                    iwn_ana_disc_noavg.newDataMFA(iwn[i]);
                    ipn_ana_disc_noavg.newDataMFA(ipn[i]);
                    iwn_ana_disc_avg.newDataMFA(iwn[i]);
                    ipn_ana_disc_avg.newDataMFA(ipn[i]);
                    iwn_ana_keep.newDataMFA(iwn[i]);
                    ipn_ana_keep.newDataMFA(ipn[i]);
                    iwn_ana_keep_no_weights.newDataMFA(iwn[i]);
                    ipn_ana_keep_no_weights.newDataMFA(ipn[i]);
                    // Compare between non-averaged and averaged version for each RESULT
                    // that is the latest hq value alone vs the latest 2^L average.
                    if (i>=firstAvgFullHistory) {
                        for (int j=0; j<20; ++j) {
                            double delta = iwn_ana_disc_noavg.hq[j] - iwn_ana_disc_avg.hq[j];
                            iwn_frame_to_frame_dev[j] += delta * delta;
                            delta = ipn_ana_disc_noavg.hq[j] - ipn_ana_disc_avg.hq[j];
                            ipn_frame_to_frame_dev[j] += delta * delta;
                        }
                    }
                }
                // finish std dev computations
                for (int j=0; j<20; ++j) {
                    iwn_frame_to_frame_dev[j] /= max_points - firstAvgFullHistory;
                    ipn_frame_to_frame_dev[j] /= max_points - firstAvgFullHistory;
                }
                
    
                double m,v;
                cout << "Results for nlevel=" << nlevel << ", nmin=" << nmin << endl;
                cout << "h(q)\tiwn_disc (dev)         \tiwn_keep (delta_no_w)    \tipn_disc (dev)         \tipn_keep (delta_no_w)"<<endl;
                cout.setf(ios_base::left);
                for (int i=0; i<20; ++i) {
                    if (i>=10) cout << (9-i) << "\t";
                    else cout << (i+1) << "\t";
                    cout << setw(8)   << iwn_ana_disc_avg.hq[i] << " ("<<setw(10)   << iwn_frame_to_frame_dev[i] <<")\t";
                    cout << setw(8)   << iwn_ana_keep.hq[i] << " (";
                    double delta_iwn = fabs(iwn_ana_keep.hq[i] - iwn_ana_keep_no_weights.hq[i]);
                    cout << setw(10)   << delta_iwn << ")    \t";
                    cout << setw(8)   << ipn_ana_disc_avg.hq[i] << " ("<<setw(10)   << ipn_frame_to_frame_dev[i]<< ")\t";
                    cout << setw(8)   << ipn_ana_keep.hq[i] << " (";
                    double delta_ipn = fabs(ipn_ana_keep.hq[i] - ipn_ana_keep_no_weights.hq[i]);
                    cout << setw(10)   << delta_ipn << ")";
                    cout << endl;

                    avg_iwn_disc[measureIdx][i]  += iwn_ana_disc_avg.hq[i];
                    dev_iwn_disc[measureIdx][i]  += iwn_ana_disc_avg.hq[i] * iwn_ana_disc_avg.hq[i];
                    avg_iwn_dev[measureIdx][i]   += iwn_frame_to_frame_dev[i];
                    dev_iwn_dev[measureIdx][i]   += iwn_frame_to_frame_dev[i] * iwn_frame_to_frame_dev[i];
                    avg_iwn_keep[measureIdx][i]  += iwn_ana_keep.hq[i];
                    dev_iwn_keep[measureIdx][i]  += iwn_ana_keep.hq[i] * iwn_ana_keep.hq[i];
                    avg_iwn_delta[measureIdx][i] += delta_iwn;
                    dev_iwn_delta[measureIdx][i] += delta_iwn * delta_iwn;

                    avg_ipn_disc[measureIdx][i]  += ipn_ana_disc_avg.hq[i];
                    dev_ipn_disc[measureIdx][i]  += ipn_ana_disc_avg.hq[i] * ipn_ana_disc_avg.hq[i];
                    avg_ipn_dev[measureIdx][i]   += ipn_frame_to_frame_dev[i];
                    dev_ipn_dev[measureIdx][i]   += ipn_frame_to_frame_dev[i] * ipn_frame_to_frame_dev[i];
                    avg_ipn_keep[measureIdx][i]  += ipn_ana_keep.hq[i];
                    dev_ipn_keep[measureIdx][i]  += ipn_ana_keep.hq[i] * ipn_ana_keep.hq[i];
                    avg_ipn_delta[measureIdx][i] += delta_ipn;
                    dev_ipn_delta[measureIdx][i] += delta_ipn * delta_ipn;
                }

                ++measureIdx;
            }
    
            // Once loops are done, we can dispose of this filter
            filters->unref();
        }
    
        delete [] iwn;
        delete [] ipn;
    }

    // Final stats
    int measureIdx = 0;
    for (int nlevel = 4; nlevel <= max_level; nlevel+=3) {
        for (int nmin = 10; nmin<=max_nmin; nmin += 20) {
            for (int i=0; i<20; ++i) {
                avg_iwn_disc[measureIdx][i]  /= Nruns;
                dev_iwn_disc[measureIdx][i]  = sqrt(dev_iwn_disc[measureIdx][i]/Nruns - avg_iwn_disc[measureIdx][i] * avg_iwn_disc[measureIdx][i]);
                avg_iwn_dev[measureIdx][i]   /= Nruns;
                dev_iwn_dev[measureIdx][i]   = sqrt(dev_iwn_dev[measureIdx][i]/Nruns - avg_iwn_dev[measureIdx][i] * avg_iwn_dev[measureIdx][i]);
                avg_iwn_keep[measureIdx][i]  /= Nruns;
                dev_iwn_keep[measureIdx][i]  = sqrt(dev_iwn_keep[measureIdx][i]/Nruns - avg_iwn_keep[measureIdx][i] * avg_iwn_keep[measureIdx][i]);
                avg_iwn_delta[measureIdx][i] /= Nruns;
                dev_iwn_delta[measureIdx][i] = sqrt(dev_iwn_delta[measureIdx][i]/Nruns - avg_iwn_delta[measureIdx][i] * avg_iwn_delta[measureIdx][i]);


                avg_ipn_disc[measureIdx][i]  /= Nruns;
                dev_ipn_disc[measureIdx][i]  = sqrt(dev_ipn_disc[measureIdx][i]/Nruns - avg_ipn_disc[measureIdx][i] * avg_ipn_disc[measureIdx][i]);
                avg_ipn_dev[measureIdx][i]   /= Nruns;
                dev_ipn_dev[measureIdx][i]   = sqrt(dev_ipn_dev[measureIdx][i]/Nruns - avg_ipn_dev[measureIdx][i] * avg_ipn_dev[measureIdx][i]);
                avg_ipn_keep[measureIdx][i]  /= Nruns;
                dev_ipn_keep[measureIdx][i]  = sqrt(dev_ipn_keep[measureIdx][i]/Nruns - avg_ipn_keep[measureIdx][i] * avg_ipn_keep[measureIdx][i]);
                avg_ipn_delta[measureIdx][i] /= Nruns;
                dev_ipn_delta[measureIdx][i] = sqrt(dev_ipn_delta[measureIdx][i]/Nruns - avg_ipn_delta[measureIdx][i] * avg_ipn_delta[measureIdx][i]);
                
            }
            
            cout << "Overall stats for nlevel=" << nlevel << ", nmin=" << nmin << ". Mean values." <<endl;
            cout << "h(q)\tiwn_disc (dev)         \tiwn_keep (delta_no_w)    \tipn_disc (dev)         \tipn_keep (delta_no_w)"<<endl;
            for (int i=0; i<20; ++i) {
                if (i>=10) cout << (9-i) << "\t";
                else cout << (i+1) << "\t";
                cout << setw(8)   << avg_iwn_disc[measureIdx][i] << " ("<<setw(10)   << avg_iwn_dev[measureIdx][i] <<")\t";
                cout << setw(8)   << avg_iwn_keep[measureIdx][i] << " (";
                cout << setw(10)   << avg_iwn_delta[measureIdx][i] << ")    \t";
                
                cout << setw(8)   << avg_ipn_disc[measureIdx][i] << " ("<<setw(10)   << avg_ipn_dev[measureIdx][i] << ")\t";
                cout << setw(8)   << avg_ipn_keep[measureIdx][i] << " (";
                cout << setw(10)   << avg_ipn_delta[measureIdx][i] << ")";
                cout << endl;
            }
            
            cout << "Overall stats for nlevel=" << nlevel << ", nmin=" << nmin << ". Standard deviations." <<endl;
            cout << "h(q)\tiwn_disc (dev)         \tiwn_keep (delta_no_w)    \tipn_disc (dev)         \tipn_keep (delta_no_w)"<<endl;
            for (int i=0; i<20; ++i) {
                if (i>=10) cout << (9-i) << "\t";
                else cout << (i+1) << "\t";
                cout << setw(8)   << dev_iwn_disc[measureIdx][i] << " ("<<setw(10)   << dev_iwn_dev[measureIdx][i] <<")\t";
                cout << setw(8)   << dev_iwn_keep[measureIdx][i] << " (";
                cout << setw(10)   << dev_iwn_delta[measureIdx][i] << ")    \t";
                
                cout << setw(8)   << dev_ipn_disc[measureIdx][i] << " ("<<setw(10)   << dev_ipn_dev[measureIdx][i] << ")\t";
                cout << setw(8)   << dev_ipn_keep[measureIdx][i] << " (";
                cout << setw(10)   << dev_ipn_delta[measureIdx][i] << ")";
                cout << endl;
            }
            
            ++measureIdx;
        }
    }
}
